#
# Copyright (c) 2006-2012, Prometheus Research, LLC
#


from . import connect, domain, introspect, split_sql, tr
from htsql.core.addon import Addon


class EngineMSSQLAddon(Addon):

    name = 'engine.mssql'
    hint = """implements HTSQL for Microsoft SQL Server"""
    help = """
    This extension implements HTSQL for Microsoft SQL Server.
    Currently MS SQL Server 2005 and 2008 are supported.

    This extension is loaded automatically when the engine of
    the database URI is set to `mssql`.
    """
    packages = ['.', '.tr']

    def __init__(self, app, attributes):
        if app.htsql.db.engine != 'mssql':
            raise ImportError("mssql engine is expected")
        super(EngineMSSQLAddon, self).__init__(app, attributes)





#
# Copyright (c) 2006-2012, Prometheus Research, LLC
#


from htsql.core.domain import (Domain, BooleanDomain, IntegerDomain,
                               DecimalDomain, FloatDomain, StringDomain,
                               DateDomain, TimeDomain, DateTimeDomain,
                               OpaqueDomain)


class MSSQLDomain(Domain):
    """
    Represents a native data type of MS SQL Server.

    This is an abstract mixin class; see subclasses for concrete data types.

    `schema_name` (a Unicode string)
        The name of the type schema.

    `name` (a Unicode string)
        The name of the type.
    """

    def __init__(self, schema_name, name, **attributes):
        # Sanity check on the arguments.
        assert isinstance(schema_name, unicode)
        assert isinstance(name, unicode)

        # Pass the attributes to the concrete domain constructor.
        super(MSSQLDomain, self).__init__(**attributes)
        self.schema_name = schema_name
        self.name = name

    def __str__(self):
        return "%s.%s" % (self.schema_name.encode('utf-8'),
                          self.name.encode('utf-8'))

    def __eq__(self, other):
        # The generic domain comparison checks if the types of the domains
        # and all their attributes are equal.  Since we added extra attributes,
        # we need to update the implementation.
        return (super(MSSQLDomain, self).__eq__(other) and
                self.schema_name == other.schema_name and
                self.name == other.name)


class MSSQLBooleanDomain(MSSQLDomain, BooleanDomain):
    """
    Represents a ``BIT`` type of MS SQL Server.

    Boolean values are expressed as integers; ``0`` is the FALSE
    value, ``1`` is the TRUE value.
    """


class MSSQLIntegerDomain(MSSQLDomain, IntegerDomain):
    """
    Represents ``SMALLINT``, ``INT``, and ``BIGINT`` data types.
    """


class MSSQLDecimalDomain(MSSQLDomain, DecimalDomain):
    """
    Represents ``DECIMAL`` and ``NUMERIC`` data types.
    """


class MSSQLFloatDomain(MSSQLDomain, FloatDomain):
    """
    Represents ``REAL`` and ``FLOAT`` data types.
    """


class MSSQLStringDomain(MSSQLDomain, StringDomain):
    """
    Represents ``CHAR``, ``VARCHAR``, ``NCHAR``, ``NVARCHAR`` data types.
    """


class MSSQLDateDomain(MSSQLDomain, DateDomain):
    """
    Represents ``DATETIME`` and ``SMALLDATETIME`` data types.
    """


class MSSQLTimeDomain(MSSQLDomain, TimeDomain):
    pass


class MSSQLDateTimeDomain(MSSQLDomain, DateTimeDomain):
    pass


class MSSQLOpaqueDomain(MSSQLDomain, OpaqueDomain):
    """
    Represents an unsupported MS SQL Server data type.
    """





#
# Copyright (c) 2006-2012, Prometheus Research, LLC
#


from htsql.core.adapter import adapts
from htsql.core.domain import (BooleanDomain, StringDomain, IntegerDomain,
                               DecimalDomain, DateDomain, TimeDomain,
                               DateTimeDomain)
from htsql.core.tr.error import SerializeError
from htsql.core.tr.frame import ColumnPhrase, ReferencePhrase, LiteralPhrase
from htsql.core.tr.dump import (FormatName, DumpBranch, DumpBySignature,
                                DumpFromPredicate, DumpToPredicate,
                                DumpIsTotallyEqual, DumpBoolean, DumpInteger,
                                DumpDecimal, DumpFloat, DumpDate, DumpTime,
                                DumpDateTime, DumpToInteger,
                                DumpToFloat, DumpToDecimal, DumpToString,
                                DumpToDate, DumpToTime, DumpToDateTime)
from htsql.core.tr.fn.dump import (DumpRound, DumpRoundTo, DumpTrunc,
                                   DumpTruncTo, DumpLength, DumpConcatenate,
                                   DumpSubstring, DumpTrim, DumpToday, DumpNow,
                                   DumpExtractYear, DumpExtractMonth,
                                   DumpExtractDay, DumpExtractHour,
                                   DumpExtractMinute, DumpExtractSecond,
                                   DumpMakeDate, DumpMakeDateTime,
                                   DumpDateIncrement, DumpDateDecrement,
                                   DumpDateTimeIncrement,
                                   DumpDateTimeDecrement, DumpDateDifference)
from htsql.core.tr.fn.signature import SortDirectionSig
from .signature import RowNumberSig


class MSSQLFormatName(FormatName):

    def __call__(self):
        self.stream.write(u"[%s]" % self.value.replace(u"]", u"]]"))


class MSSQLDumpBranch(DumpBranch):

    def dump_select(self):
        aliases = self.state.select_aliases_by_tag[self.frame.tag]
        self.write(u"SELECT ")
        self.indent()
        if self.frame.limit is not None:
            self.write(u"TOP "+unicode(self.frame.limit))
            self.newline()
        for index, phrase in enumerate(self.frame.select):
            alias = None
            if self.state.hook.with_aliases:
                alias = aliases[index]
                if isinstance(phrase, ColumnPhrase):
                    if alias == phrase.column.name:
                        alias = None
                if isinstance(phrase, ReferencePhrase):
                    target_alias = (self.state.select_aliases_by_tag
                                            [phrase.tag][phrase.index])
                    if alias == target_alias:
                        alias = None
            if alias is not None:
                self.format("{selection} AS {alias:name}",
                            selection=phrase, alias=alias)
            else:
                self.format("{selection}",
                            selection=phrase)
            if index < len(self.frame.select)-1:
                self.write(u",")
                self.newline()
        self.dedent()

    def dump_group(self):
        if not self.frame.group:
            return
        self.newline()
        self.write(u"GROUP BY ")
        for index, phrase in enumerate(self.frame.group):
            self.format("{kernel}", kernel=phrase)
            if index < len(self.frame.group)-1:
                self.write(u", ")

    def dump_limit(self):
        assert self.frame.offset is None


class MSSQLDumpFromPredicate(DumpFromPredicate):

    def __call__(self):
        if self.phrase.is_nullable:
            self.format("(CASE WHEN {op} THEN 1 WHEN NOT {op} THEN 0 END)",
                        self.arguments)
        else:
            self.format("(CASE WHEN {op} THEN 1 ELSE 0 END)",
                        self.arguments)


class MSSQLDumpToPredicate(DumpToPredicate):

    def __call__(self):
        self.format("({op} <> 0)", self.arguments)


class MSSQLDumpRowNumber(DumpBySignature):

    adapts(RowNumberSig)

    def __call__(self):
        self.format("ROW_NUMBER() OVER (ORDER BY {ops:union{, }})",
                    self.arguments)


class MSSQLDumpSortDirection(DumpBySignature):

    adapts(SortDirectionSig)

    def __call__(self):
        self.format("{base} {direction:switch{ASC|DESC}}",
                    self.arguments, self.signature)


class MSSQLDumpBoolean(DumpBoolean):

    def __call__(self):
        if self.value is True:
            self.write(u"1")
        if self.value is False:
            self.write(u"0")


class MSSQLDumpInteger(DumpInteger):

    def __call__(self):
        if not (-2**63 <= self.value < 2**63):
            raise SerializeError("invalid integer value",
                                 self.phrase.mark)
        if abs(self.value) < 2**31:
            if self.value >= 0:
                self.write(unicode(self.value))
            else:
                self.write(u"(%s)" % self.value)
        else:
            self.write(u"CAST(%s AS BIGINT)" % self.value)


class MSSQLDumpFloat(DumpFloat):

    def __call__(self):
        assert str(self.value) not in ['inf', '-inf', 'nan']
        value = repr(self.value)
        if 'e' not in value and 'E' not in value:
            value = value+'e0'
        if value[0] == '-':
            value = "(%s)" % value
        self.write(unicode(value))


class MSSQLDumpDecimal(DumpDecimal):

    def __call__(self):
        assert self.value.is_finite()
        value = str(self.value)
        if 'E' in value:
            value = "CAST(%s AS DECIMAL(38,19))" % value
        elif '.' not in value:
            value = "%s." % value
        if value[0] == '-':
            value = "(%s)" % value
        self.write(unicode(value))


class MSSQLDumpDate(DumpDate):

    def __call__(self):
        self.format("CAST({value:literal} AS DATETIME)",
                    value=unicode(self.value))


class MSSQLDumpTime(DumpTime):

    def __call__(self):
        value = (self.value.hour*3600 + self.value.minute*60 +
                 self.value.second + self.value.microsecond/1000000.0) / 86400.0
        value = repr(value)
        if 'e' not in value and 'E' not in value:
            value = value+'e0'
        self.write(unicode(value))


class MSSQLDumpDateTime(DumpDateTime):

    def __call__(self):
        value = self.value.replace(tzinfo=None)
        if not value.microsecond:
            value = unicode(value)
        else:
            value = unicode(value)[:-3]
        self.format("CAST({value:literal} AS DATETIME)", value=value)


class MSSQLDumpToInteger(DumpToInteger):

    def __call__(self):
        self.format("CAST({base} AS INT)", base=self.base)


class MSSQLDumpToFloat(DumpToFloat):

    def __call__(self):
        self.format("CAST({base} AS FLOAT)", base=self.base)


class MSSQLDumpToDecimal(DumpToDecimal):

    def __call__(self):
        self.format("CAST({base} AS DECIMAL(38,19))", base=self.base)


class MSSQLDumpIntegerToDecimal(MSSQLDumpToDecimal):

    adapts(IntegerDomain, DecimalDomain)

    def __call__(self):
        self.format("CAST({base} AS DECIMAL(38))", base=self.base)


class MSSQLDumpToString(DumpToString):

    def __call__(self):
        self.format("CAST({base} AS VARCHAR(MAX))", base=self.base)


class MSSQLDumpBooleanToString(DumpToString):

    adapts(BooleanDomain, StringDomain)

    def __call__(self):
        if self.base.is_nullable:
            self.format("(CASE WHEN {base} <> 0 THEN 'true'"
                        " WHEN NOT {base} = 0 THEN 'false' END)",
                        base=self.base)
        else:
            self.format("(CASE WHEN {base} <> 0 THEN 'true' ELSE 'false' END)",
                        base=self.base)


class MSSQLDumpDateToString(DumpToString):

    adapts(DateDomain, StringDomain)

    def __call__(self):
        self.format("SUBSTRING(CONVERT(VARCHAR, {base}, 21), 1, 10)",
                    base=self.base)


class MSSQLDumpTimeToString(DumpToString):

    adapts(TimeDomain, StringDomain)

    def __call__(self):
        self.format("SUBSTRING(CONVERT(VARCHAR, CAST({base} AS DATETIME), 21),"
                    " 12, 12)", base=self.base)


class MSSQLDumpDateTimeToString(DumpToString):

    adapts(DateTimeDomain, StringDomain)

    def __call__(self):
        self.format("CONVERT(VARCHAR, {base}, 21)", base=self.base)


class MSSQLDumpStringToDate(DumpToDate):

    adapts(StringDomain, DateDomain)

    def __call__(self):
        self.format("CAST(FLOOR(CAST(CAST({base} AS DATETIME) AS FLOAT))"
                    " AS DATETIME)", base=self.base)


class MSSQLDumpDateTimeToDate(DumpToDate):

    adapts(DateTimeDomain, DateDomain)

    def __call__(self):
        self.format("CAST(FLOOR(CAST({base} AS FLOAT)) AS DATETIME)",
                    base=self.base)


class MSSQLDumpStringToTime(DumpToTime):

    adapts(StringDomain, TimeDomain)

    def __call__(self):
        self.format("CAST(CAST('1900-01-01 ' + {base} AS DATETIME) AS FLOAT)",
                    base=self.base)


class MSSQLDumpDateTimeToTime(DumpToTime):

    adapts(DateTimeDomain, TimeDomain)

    def __call__(self):
        self.format("(CAST({base} AS FLOAT) - FLOOR(CAST({base} AS FLOAT)))",
                    base=self.base)


class MSSQLDumpStringToDateTime(DumpToDateTime):

    adapts(StringDomain, DateTimeDomain)

    def __call__(self):
        self.format("CAST({base} AS DATETIME)", base=self.base)


class MSSQLDumpDateToDateTime(DumpToDateTime):

    adapts(DateDomain, DateTimeDomain)

    def __call__(self):
        self.format("{base}", base=self.base)


class MSSQLDumpIsTotallyEqual(DumpIsTotallyEqual):

    def __call__(self):
        self.format("((CASE WHEN ({lop} = {rop}) OR"
                    " ({lop} IS NULL AND {rop} IS NULL)"
                    " THEN 1 ELSE 0 END) {polarity:switch{<>|=}} 0)",
                    self.arguments, self.signature)


class MSSQLDumpRound(DumpRound):

    def __call__(self):
        if isinstance(self.phrase.domain, DecimalDomain):
            self.format("CAST(ROUND({op}, 0) AS DECIMAL(38))", self.arguments)
        else:
            self.format("ROUND({op}, 0)", self.arguments)


class MSSQLDumpRoundTo(DumpRoundTo):

    def __call__(self):
        scale = None
        if (isinstance(self.phrase.precision, LiteralPhrase) and
            self.phrase.precision.value is not None):
            scale = self.phrase.precision.value
            if scale < 0:
                scale = 0
        if scale is not None:
            self.format("CAST(ROUND({op}, {precision})"
                        " AS DECIMAL(38,{scale:pass}))",
                        self.arguments, self.signature, scale=unicode(scale))
        else:
            self.format("ROUND({op}, {precision})",
                        self.arguments, self.signature)


class MSSQLDumpTrunc(DumpTrunc):

    def __call__(self):
        if isinstance(self.phrase.domain, DecimalDomain):
            self.format("CAST(ROUND({op} -"
                        " (CASE WHEN {op} >= 0 THEN 0.5 ELSE -0.5 END), 0)"
                        " AS DECIMAL(38))", self.arguments)
        else:
            self.format("ROUND({op} -"
                        " (CASE WHEN {op} >= 0 THEN 0.5 ELSE -0.5 END), 0)",
                        self.arguments)


class MSSQLDumpTruncTo(DumpTruncTo):

    def __call__(self):
        scale = None
        if (isinstance(self.phrase.precision, LiteralPhrase) and
            self.phrase.precision.value is not None):
            scale = self.phrase.precision.value
            if scale < 0:
                scale = 0
        if scale is not None:
            self.format("CAST(ROUND({op} -"
                        " (CASE WHEN {op} >= 0 THEN 0.5 ELSE -0.5 END) /"
                        " CAST(POWER(10e0, {precision}) AS DECIMAL(38,19)),"
                        " {precision})"
                        " AS DECIMAL(38,{scale:pass}))",
                        self.arguments, self.signature, scale=unicode(scale))
        else:
            self.format("ROUND({op} -"
                        " (CASE WHEN {op} >= 0 THEN 0.5 ELSE -0.5 END) /"
                        " CAST(POWER(10e0, {precision}) AS DECIMAL(38,19)),"
                        " {precision})", self.arguments, self.signature)


class MSSQLDumpLength(DumpLength):

    template = "LEN({op})"


class MSSQLDumpConcatenate(DumpConcatenate):

    template = "({lop} + {rop})"


class MSSQLDumpSubstring(DumpSubstring):

    def __call__(self):
        if self.phrase.length is None:
            self.format("SUBSTRING({op}, {start}, LEN({op}))", self.phrase)
        else:
            self.format("SUBSTRING({op}, {start}, {length})", self.phrase)


class MSSQLDumpTrim(DumpTrim):

    def __call__(self):
        if self.signature.is_left and not self.signature.is_right:
            self.format("LTRIM({op})", self.arguments)
        elif not self.signature.is_left and self.signature.is_right:
            self.format("RTRIM({op})", self.arguments)
        else:
            self.format("LTRIM(RTRIM({op}))", self.arguments)


class MSSQLDumpToday(DumpToday):

    template = "CAST(FLOOR(CAST(GETDATE() AS FLOAT)) AS DATETIME)"


class MSSQLDumpNow(DumpNow):

    template = "GETDATE()"


class MSSQLDumpExtractYear(DumpExtractYear):

    template = "DATEPART(YEAR, {op})"


class MSSQLDumpExtractMonth(DumpExtractMonth):

    template = "DATEPART(MONTH, {op})"


class MSSQLDumpExtractDay(DumpExtractDay):

    template = "DATEPART(DAY, {op})"


class MSSQLDumpExtractHour(DumpExtractHour):

    template = "DATEPART(HOUR, {op})"


class MSSQLDumpExtractMinute(DumpExtractMinute):

    template = "DATEPART(MINUTE, {op})"


class MSSQLDumpExtractSecond(DumpExtractSecond):

    template = ("(DATEPART(SECOND, {op}) +"
                " DATEPART(MILLISECOND, {op}) / 1000e0)")


class MSSQLDumpMakeDate(DumpMakeDate):

    template = ("DATEADD(DAY, {day} - 1,"
                " DATEADD(MONTH, {month} - 1,"
                " DATEADD(YEAR, {year} - 2001,"
                " CAST('2001-01-01' AS DATETIME))))")


class MSSQLDumpMakeDateTime(DumpMakeDateTime):

    def __call__(self):
        template = ("DATEADD(DAY, {day} - 1,"
                    " DATEADD(MONTH, {month} - 1,"
                    " DATEADD(YEAR, {year} - 2001,"
                    " CAST('2001-01-01' AS DATETIME))))")
        if self.phrase.hour is not None:
            template = "DATEADD(HOUR, {hour}, %s)" % template
        if self.phrase.minute is not None:
            template = "DATEADD(MINUTE, {minute}, %s)" % template
        if self.phrase.second is not None:
            template = "DATEADD(MILLISECOND, 1000 * {second}, %s)" % template
        self.format(template, self.arguments)


class MSSQLDumpDateIncrement(DumpDateIncrement):

    template = "({lop} + {rop})"


class MSSQLDumpDateTimeIncrement(DumpDateTimeIncrement):

    template = "({lop} + {rop})"


class MSSQLDumpDateDecrement(DumpDateDecrement):

    template = "({lop} - {rop})"


class MSSQLDumpDateTimeDecrement(DumpDateTimeDecrement):

    template = "({lop} - {rop})"


class MSSQLDumpDateDifference(DumpDateDifference):

    template = "DATEDIFF(DAY, {rop}, {lop})"





#
# Copyright (c) 2006-2012, Prometheus Research, LLC
#


from htsql.core.tr.flow import LiteralCode, FormulaCode
from htsql.core.tr.fn.signature import ReplaceSig, ConcatenateSig, LikeSig
from htsql.core.tr.fn.encode import EncodeContains


class MSSQLEncodeContains(EncodeContains):

    def __call__(self):
        lop = self.state.encode(self.binding.lop)
        rop = self.state.encode(self.binding.rop)
        if isinstance(rop, LiteralCode):
            if rop.value is not None:
                value = (u"%" + rop.value.replace(u"\\", u"\\\\")
                                         .replace(u"[", u"\\[")
                                         .replace(u"]", u"\\]")
                                         .replace(u"%", u"\\%")
                                         .replace(u"_", u"\\_") + u"%")
                rop = rop.clone(value=value)
        else:
            backslash_literal = LiteralCode(u"\\", rop.domain, self.binding)
            xbackslash_literal = LiteralCode(u"\\\\", rop.domain, self.binding)
            lbracket_literal = LiteralCode(u"[", rop.domain, self.binding)
            xlbracket_literal = LiteralCode(u"\\[", rop.domain, self.binding)
            rbracket_literal = LiteralCode(u"]", rop.domain, self.binding)
            xrbracket_literal = LiteralCode(u"\\]", rop.domain, self.binding)
            percent_literal = LiteralCode(u"%", rop.domain, self.binding)
            xpercent_literal = LiteralCode(u"\\%", rop.domain, self.binding)
            underscore_literal = LiteralCode(u"_", rop.domain, self.binding)
            xunderscore_literal = LiteralCode(u"\\_", rop.domain, self.binding)
            rop = FormulaCode(ReplaceSig(), rop.domain, self.binding,
                              op=rop, old=backslash_literal,
                              new=xbackslash_literal)
            rop = FormulaCode(ReplaceSig(), rop.domain, self.binding,
                              op=rop, old=lbracket_literal,
                              new=xlbracket_literal)
            rop = FormulaCode(ReplaceSig(), rop.domain, self.binding,
                              op=rop, old=rbracket_literal,
                              new=xrbracket_literal)
            rop = FormulaCode(ReplaceSig(), rop.domain, self.binding,
                              op=rop, old=percent_literal,
                              new=xpercent_literal)
            rop = FormulaCode(ReplaceSig(), rop.domain, self.binding,
                              op=rop, old=underscore_literal,
                              new=xunderscore_literal)
            rop = FormulaCode(ConcatenateSig(), rop.domain, self.binding,
                              lop=percent_literal, rop=rop)
            rop = FormulaCode(ConcatenateSig(), rop.domain, self.binding,
                              lop=rop, rop=percent_literal)
        return FormulaCode(self.signature.clone_to(LikeSig),
                           self.domain, self.binding, lop=lop, rop=rop)





#
# Copyright (c) 2006-2012, Prometheus Research, LLC
#


from htsql.core.connect import Connect, Normalize, NormalizeError, DBError
from htsql.core.adapter import adapts
from htsql.core.context import context
from htsql.core.domain import (BooleanDomain, StringDomain, DateDomain,
                               TimeDomain)
import datetime
import pymssql


class MSSQLError(DBError):
    """
    Raised when a database error occurred.
    """


class ConnectMSSQL(Connect):

    def open(self):
        db = context.app.htsql.db
        parameters = {}
        parameters['database'] = db.database
        if db.host is not None:
            if db.port is not None:
                parameters['host'] = "%s:%s" % (db.host, db.port)
            else:
                parameters['host'] = db.host
        if db.username is not None:
            parameters['user'] = db.username
        if db.password is not None:
            parameters['password'] = db.password
        parameters['charset'] = 'utf8'
        connection = pymssql.connect(**parameters)
        if self.with_autocommit:
            connection.autocommit(True)
        return connection


class NormalizeMSSQLError(NormalizeError):

    def __call__(self):
        # If we got a DBAPI exception, generate our error out of it.
        if isinstance(self.error, pymssql.Error):
            message = str(self.error)
            error = MSSQLError(message)
            return error

        # Otherwise, let the superclass return `None`.
        return super(NormalizeMSSQLError, self).__call__()


class NormalizeMSSQLBoolean(Normalize):

    adapts(BooleanDomain)

    @staticmethod
    def convert(value):
        if value is None:
            return None
        return (value != 0)


class NormalizeMSSQLString(Normalize):

    adapts(StringDomain)

    @staticmethod
    def convert(value):
        if isinstance(value, str):
            value = value.decode('utf-8')
        return value


class NormalizeMSSQLDate(Normalize):

    adapts(DateDomain)

    @staticmethod
    def convert(value):
        if isinstance(value, datetime.datetime):
            assert not value.time()
            value = value.date()
        return value


class NormalizeMSSQLTime(Normalize):

    adapts(TimeDomain)

    @staticmethod
    def convert(value):
        if isinstance(value, float):
            assert 0.0 <= value < 1.0
            value = int(86400000000*value) * datetime.timedelta(0,0,1)
            assert not value.days
            value = (datetime.datetime(2001,1,1) + value).time()
        return value





#
# Copyright (c) 2006-2012, Prometheus Research, LLC
#


from htsql.core.adapter import Protocol, named
from htsql.core.introspect import Introspect
from htsql.core.entity import make_catalog
from .domain import (MSSQLBooleanDomain, MSSQLIntegerDomain,
                     MSSQLDecimalDomain, MSSQLFloatDomain, MSSQLStringDomain,
                     MSSQLDateTimeDomain, MSSQLOpaqueDomain)
from htsql.core.connect import connect
import itertools
import fnmatch


class IntrospectMSSQL(Introspect):

    system_schema_names = [u'guest', u'INFORMATION_SCHEMA', u'sys', u'db_*']

    def __call__(self):
        connection = connect()
        cursor = connection.cursor()

        catalog = make_catalog()

        schema_by_id = {}
        cursor.execute("""
            SELECT schema_id, name
            FROM sys.schemas
            ORDER BY name
        """)
        for row in cursor.fetchnamed():
            if any(fnmatch.fnmatchcase(row.name, pattern)
                   for pattern in self.system_schema_names):
                continue
            schema = catalog.add_schema(row.name)
            schema_by_id[row.schema_id] = schema

        cursor.execute("""
            SELECT default_schema_name
            FROM sys.database_principals
            WHERE principal_id = USER_ID()
        """)
        default_schema_name = cursor.fetchone()[0]
        if default_schema_name in catalog.schemas:
            catalog.schemas[default_schema_name].set_priority(1)

        table_by_id = {}
        cursor.execute("""
            SELECT object_id, schema_id, name
            FROM sys.objects
            WHERE type in ('U', 'V')
            ORDER BY schema_id, name
        """)
        for row in cursor.fetchnamed():
            if row.schema_id not in schema_by_id:
                continue
            schema = schema_by_id[row.schema_id]
            table = schema.add_table(row.name)
            table_by_id[row.object_id] = table

        column_by_id = {}
        cursor.execute("""
            SELECT c.object_id, c.column_id, c.name, c.max_length,
                   c.precision, c.scale, c.is_nullable, c.default_object_id,
                   t.name AS type_name, s.name AS type_schema_name
            FROM sys.columns c
            JOIN sys.types t ON (c.user_type_id = t.user_type_id)
            JOIN sys.schemas s ON (t.schema_id = s.schema_id)
            ORDER BY c.object_id, c.column_id
        """)
        for row in cursor.fetchnamed():
            if row.object_id not in table_by_id:
                continue
            table = table_by_id[row.object_id]
            name = row.name
            type_schema_name = row.type_schema_name
            type_name = row.type_name
            length = row.max_length if row.max_length != -1 else None
            precision = row.precision
            scale = row.scale
            introspect_domain = IntrospectMSSQLDomain(type_schema_name,
                                                      type_name,
                                                      length, precision, scale)
            domain = introspect_domain()
            is_nullable = bool(row.is_nullable)
            has_default = bool(row.default_object_id)
            column = table.add_column(name, domain, is_nullable, has_default)
            column_by_id[row.object_id, row.column_id] = column

        cursor.execute("""
            SELECT object_id, index_id, is_primary_key, is_unique_constraint
            FROM sys.indexes
            WHERE (is_primary_key = 1 OR is_unique_constraint = 1) AND
                  is_disabled = 0
            ORDER BY object_id, index_id
        """)
        index_rows = cursor.fetchnamed()
        cursor.execute("""
            SELECT object_id, index_id, index_column_id, column_id
            FROM sys.index_columns
            ORDER BY object_id, index_id, index_column_id
        """)
        column_rows_by_id = \
                dict((key, list(group))
                     for key, group in itertools.groupby(cursor.fetchnamed(),
                                         lambda r: (r.object_id, r.index_id)))
        for row in index_rows:
            if row.object_id not in table_by_id:
                continue
            table = table_by_id[row.object_id]
            key = (row.object_id, row.index_id)
            if key not in column_rows_by_id:
                continue
            column_rows = column_rows_by_id[key]
            if not all((column_row.object_id, column_row.column_id)
                            in column_by_id
                       for column_row in column_rows):
                continue
            columns = [column_by_id[column_row.object_id, column_row.column_id]
                       for column_row in column_rows]
            is_primary = bool(row.is_primary_key)
            table.add_unique_key(columns, is_primary)

        cursor.execute("""
            SELECT object_id, parent_object_id, referenced_object_id
            FROM sys.foreign_keys
            WHERE is_disabled = 0
            ORDER BY object_id
        """)
        key_rows = cursor.fetchnamed()
        cursor.execute("""
            SELECT constraint_object_id, constraint_column_id,
                   parent_object_id, parent_column_id,
                   referenced_object_id, referenced_column_id
            FROM sys.foreign_key_columns
            ORDER BY constraint_object_id, constraint_column_id
        """)
        key_column_rows_by_id = \
                dict((key, list(group))
                     for key, group in itertools.groupby(cursor.fetchnamed(),
                                            lambda r: r.constraint_object_id))
        for row in key_rows:
            if row.parent_object_id not in table_by_id:
                continue
            table = table_by_id[row.parent_object_id]
            if row.referenced_object_id not in table_by_id:
                continue
            target_table = table_by_id[row.referenced_object_id]
            if row.object_id not in key_column_rows_by_id:
                continue
            column_rows = key_column_rows_by_id[row.object_id]
            column_ids = [(column_row.parent_object_id,
                           column_row.parent_column_id)
                          for column_row in column_rows]
            target_column_ids = [(column_row.referenced_object_id,
                                  column_row.referenced_column_id)
                                 for column_row in column_rows]
            if not all(column_id in column_by_id
                       for column_id in column_ids):
                continue
            columns = [column_by_id[column_id]
                       for column_id in column_ids]
            if not all(column_id in column_by_id
                       for column_id in target_column_ids):
                continue
            target_columns = [column_by_id[column_id]
                              for column_id in target_column_ids]
            table.add_foreign_key(columns, target_table, target_columns)

        connection.release()
        return catalog


class IntrospectMSSQLDomain(Protocol):

    @classmethod
    def dispatch(component, schema_name, type_name, *args, **kwds):
        return (schema_name.encode('utf-8'), type_name.encode('utf-8'))

    @classmethod
    def matches(component, dispatch_key):
        return (dispatch_key in component.names)

    def __init__(self, schema_name, type_name,
                 length, precision, scale):
        self.schema_name = schema_name
        self.type_name = type_name
        self.length = length
        self.precision = precision
        self.scale = scale

    def __call__(self):
        return MSSQLOpaqueDomain(self.schema_name, self.type_name)


class IntrospectMSSQLCharDomain(IntrospectMSSQLDomain):

    named(('sys', 'char'), ('sys', 'nchar'))

    def __call__(self):
        return MSSQLStringDomain(self.schema_name, self.type_name,
                                 length=self.length, is_varying=False)


class IntrospectMSSQLVarCharDomain(IntrospectMSSQLDomain):

    named(('sys', 'varchar'), ('sys', 'nvarchar'))

    def __call__(self):
        return MSSQLStringDomain(self.schema_name, self.type_name,
                                 length=self.length, is_varying=False)


class IntrospectMSSQLBitDomain(IntrospectMSSQLDomain):

    named(('sys', 'bit'))

    def __call__(self):
        return MSSQLBooleanDomain(self.schema_name, self.type_name)


class IntrospectMSSQLIntegerDomain(IntrospectMSSQLDomain):

    named(('sys', 'tinyint'), ('sys', 'smallint'),
          ('sys', 'int'), ('sys', 'bigint'))

    def __call__(self):
        return MSSQLIntegerDomain(self.schema_name, self.type_name,
                                  size=self.length*8)


class IntrospectMSSQLDecimalDomain(IntrospectMSSQLDomain):

    named(('sys', 'decimal'), ('sys', 'numeric'))

    def __call__(self):
        return MSSQLDecimalDomain(self.schema_name, self.type_name,
                                  precision=self.precision, scale=self.scale)


class IntrospectMSSQLFloatDomain(IntrospectMSSQLDomain):

    named(('sys', 'real'), ('sys', 'float'))

    def __call__(self):
        return MSSQLFloatDomain(self.schema_name, self.type_name,
                                size=self.length*8)


class IntrospectMSSQLDateTimeDomain(IntrospectMSSQLDomain):

    named(('sys', 'datetime'), ('sys', 'smalldatetime'))

    def __call__(self):
        return MSSQLDateTimeDomain(self.schema_name, self.type_name)





#
# Copyright (c) 2006-2012, Prometheus Research, LLC
#


from htsql.core.split_sql import SQLToken, SplitSQL


class SplitMSSQL(SplitSQL):
    """
    Implements the SQL splitter for MS SQL Server.
    """

    tokens = [
            # Whitespace between separate statements.
            SQLToken(r"""
                     # whitespaces
                     [\ \t\r\n]+
                     # or a SQL comment
                     | -- [^\r\n]* \r?\n
                     # or a C-style comment
                     | /\* .*? \*/
                     """, only_level=0, is_junk=True),

            # The beginning of a SQL statement.
            SQLToken(r""" [a-zA-Z]+ """, only_level=0, delta=+1),

            # A block of regular SQL tokens.
            # FIXME: \-escaping for string literals.
            SQLToken(r"""
                     (
                     # whitespaces
                     [\ \t\r\n]+
                     # or a SQL comment
                     | -- [^\r\n]*\r?\n
                     # or a C-style comment
                     | /\* .*? \*/
                     # or a string literal
                     | ' (?: [^'] | '' )* '
                     # or a quoted name
                     | " (?: [^"]+ | "" )+ "
                     | \[ (?: [^\]]+ | \]\] )+ \]
                     # or a keyword or a name
                     | [a-zA-Z_][0-9a-zA-Z_]*
                     # or a number
                     | [0-9]+ (?: \. [0-9]* )? (?: [eE] [+-] [0-9]+ )?
                     # or a symbol
                     | [().,<>=!@#&|~*/%+-]
                     )+
                     """, min_level=1),

            # Semicolon at the top level indicates the statement end.
            SQLToken(r""" ; """, only_level=1, delta=-1),

            # Same for EOF, but it also stops the splitter.
            SQLToken(r""" $ """, only_level=1, delta=-1, is_end=True),

            # EOF outside the statement stops the splitter.
            SQLToken(r""" $ """, only_level=0, is_end=True),
    ]





#
# Copyright (c) 2006-2012, Prometheus Research, LLC
#


from htsql.core.tr.signature import isformula, FromPredicateSig, ToPredicateSig
from htsql.core.tr.reduce import ReduceFromPredicate, ReduceToPredicate


class MSSQLReduceFromPredicate(ReduceFromPredicate):

    def __call__(self):
        op = self.state.reduce(self.phrase.op)
        if isformula(op, ToPredicateSig):
            return op.op
        return self.phrase.clone(is_nullable=op.is_nullable, op=op)


class MSSQLReduceToPredicate(ReduceToPredicate):

    def __call__(self):
        op = self.state.reduce(self.phrase.op)
        if isformula(op, FromPredicateSig):
            return op.op
        return self.phrase.clone(is_nullable=op.is_nullable, op=op)





#
# Copyright (c) 2006-2012, Prometheus Research, LLC
#


from htsql.core.domain import BooleanDomain, IntegerDomain
from htsql.core.tr.term import PermanentTerm, FilterTerm
from htsql.core.tr.flow import LiteralCode, FormulaCode, ScalarUnit
from htsql.core.tr.coerce import coerce
from htsql.core.tr.signature import CompareSig, AndSig
from htsql.core.tr.fn.signature import SortDirectionSig
from .signature import RowNumberSig
from htsql.core.tr.compile import CompileOrdered
from htsql.core.tr.stitch import arrange, spread


class MSSQLCompileOrdered(CompileOrdered):

    def __call__(self):
        if self.flow.offset is None:
            return super(MSSQLCompileOrdered, self).__call__()
        kid = self.state.compile(self.flow.base,
                                 baseline=self.state.root)
        order = arrange(self.flow)
        kid = self.state.inject(kid, [code for code, direction in order])
        ops = []
        for code, direction in order:
            op = FormulaCode(SortDirectionSig(direction=direction),
                             code.domain, code.binding, base=code)
            ops.append(op)
        row_number_code = FormulaCode(RowNumberSig(), coerce(IntegerDomain()),
                                      self.flow.binding, ops=ops)
        row_number_unit = ScalarUnit(row_number_code, self.flow.base,
                                     self.flow.binding)
        tag = self.state.tag()
        routes = kid.routes.copy()
        routes[row_number_unit] = tag
        kid = PermanentTerm(tag, kid, kid.flow, kid.baseline, routes)
        left_limit = self.flow.offset+1
        right_limit = None
        if self.flow.limit is not None:
            right_limit = self.flow.limit+self.flow.offset+1
        left_limit_code = LiteralCode(left_limit, coerce(IntegerDomain()),
                                      self.flow.binding)
        right_limit_code = None
        if right_limit is not None:
            right_limit_code = LiteralCode(right_limit, coerce(IntegerDomain()),
                                           self.flow.binding)
        left_filter = FormulaCode(CompareSig('>='), coerce(BooleanDomain()),
                                  self.flow.binding,
                                  lop=row_number_unit, rop=left_limit_code)
        right_filter = None
        if right_limit_code is not None:
            right_filter = FormulaCode(CompareSig('<'),
                                       coerce(BooleanDomain()),
                                       self.flow.binding,
                                       lop=row_number_unit,
                                       rop=right_limit_code)
        filter = left_filter
        if right_filter is not None:
            filter = FormulaCode(AndSig(), coerce(BooleanDomain()),
                                 self.flow.binding,
                                 ops=[left_filter, right_filter])
        routes = kid.routes.copy()
        for unit in spread(self.flow):
            routes[unit] = routes[unit.clone(flow=self.backbone)]
        return FilterTerm(self.state.tag(), kid, filter,
                          self.flow, kid.baseline, routes)



